"""External methods for Tracker-related ZClasses and Products."""

import string
import OFS.Image
import urllib
import smtplib
from DocumentTemplate import html_quote
from AccessControl.PermissionRole import rolesForPermissionOn

def Tracker_text_content(self):
    """Return the text content of the object, for consumption by catalog."""
    return self.description

def addTracker(self, id, REQUEST=None):
    """ """
    # self = Factory - actually, FactoryDispatcher
    # Destination() = the folder in which the obj is being created.
    if not id:
        raise ValueError, "You must specify an id."
    inst = self.Tracker(id)

    dest = self.Destination()
    dest._setObject(id, inst)
    tracker = getattr(dest, id)

    __traceback_info__ = (`self`, `inst`, `dest`, `tracker`)

    setTrackerRules(tracker)

    user = (REQUEST and `REQUEST.AUTHENTICATED_USER`) or ''
    for role in ['TrackerOwner', 'Supporter', 'Requester', 'IssueOwner']:
        if not tracker._has_user_defined_role(role):
            tracker._addRole(role)
    tracker.manage_addLocalRoles(user, ['TrackerOwner'])

    for ps in tracker.propertysheets:
        ps.manage_changeProperties(REQUEST)

    tracker.supporters = []

    tracker.index_object()

    if REQUEST is not None:
        return REQUEST['RESPONSE'].redirect(self.DestinationURL()
                                            + '/' + id
                                            + '/configBasicForm')

def setTrackerConfig(self, REQUEST):
    """Process fundamental configuration activities.

    If request has a 'nextpage' setting, we redirect there, otherwise we go 
    to the tracker proper."""

    self.unindex_object()
    if REQUEST.get('config_form', '') == 'basic':
        setTrackerConfigBasic(self, REQUEST)

    elif REQUEST.get('config_form', '') == 'traitvalues':
        setTrackerConfigTraitValues(self, REQUEST)

    elif REQUEST.get('config_form', '') == 'assignments':
        setTrackerConfigAssignments(self, REQUEST)

    else:
        raise 'Bad Request', "Unrecognized tracker configuration activity."

    for ps in self.propertysheets:
        ps.manage_changeProperties(REQUEST)

    self.index_object()

    REQUEST.set('manage_tabs_message', 'Changed!')
    REQUEST["RESPONSE"].redirect(REQUEST.get('nextpage',
                                             self.absolute_url()))

def setTrackerConfigBasic(self, REQUEST):
    """Do setTrackerConfig operations for basic config form."""

    self.did_basic = 1

    traits = REQUEST.get('traits', [])
    traits = self.traits = unique(map(string.strip, traits), noblanks=1)
    REQUEST.set('traits', traits)
    __traceback_info__ = ('traits', `traits`, `REQUEST.get('traits')`)

    # tracker.supporters: [supporter, ...]
    if hasattr(self, 'supporters'):
        were_supporters = filter(None, self.supporters)
    else:
        were_supporters = []
    are_supporters = unique(REQUEST.get('supporters', ""), noblanks=1)
    self.supporters = are_supporters
    __traceback_info__ = __traceback_info__ + ('supporters',
                                               `are_supporters`)
    jiggle_supporter_roles(self, self.supporters, were_supporters)

    # tracker.assignments: {supporter: {trait: [vals, ...]}}
    # Adjust or create supporter/trait assignments data structure.
    sa = {}  
    if hasattr(self, 'assignments'):
        sa.update(self.assignments)
    for s in are_supporters:
        # Make sure each supporter has a slot for each trait.  (Note that
        # we do not cull defunct supporters, traits, or trait values - maybe
        # they'll come back...)
        d = sa.get(s, {})
        for t in traits:
            if not d.has_key(t):
                d[t] = (t + "_range", [])
        sa[s] = d
    self.assignments = sa
    __traceback_info__ = __traceback_info__ + ('assignments', `sa`)

    # tracker.traitsStruct: {range_nm: (val_nm, [vals, ...])}
    # Adjust or create user specified traits data structure
    traitsStruct = {}
    use_current = 0
    if hasattr(self, 'traitsStruct'):
        # Reconcile with prior version of traits struct.
        traitsStruct.update(self.traitsStruct)
        traitsRange = map(lambda x: x + "_range", traits)
        # Cull defunct traits:
        for i in traitsStruct.keys():
            if i not in traitsRange:
                del traitsStruct[i]
    for val_nm in traits:
        range_nm = val_nm + "_range"
        if traitsStruct.has_key(range_nm):
            continue
        elif hasattr(self, range_nm):
            traitsStruct[range_nm] = (val_nm, getattr(self, range_nm))
        else:
            traitsStruct[range_nm] = (val_nm, [])
    self.traitsStruct = traitsStruct
    __traceback_info__ = __traceback_info__ + ('traitsStruct',
                                               `traitsStruct`)

    if not hasattr(self, 'pickers'):
        pickers = {}
    else:
        pickers = self.pickers
        for i in pickers.keys():
            if i not in traits + [' __requesters']:
                del pickers[i]
    self.pickers = pickers
    __traceback_info__ = ('traits', `self.traits`, `REQUEST.get('traits')`)

def setTrackerConfigTraitValues(self, REQUEST):
    """Adjust traitsStructure according to trait/values config form."""

    self.did_traitvalues = 1

    __traceback_info__ = ('pickers', `self.pickers`)
    traits = self.traits

    traitsStruct = self.traitsStruct
    for val_nm in self.traits:
        range_nm = val_nm + "_range"
        newvals = REQUEST.get(range_nm, None)
        if newvals is not None:
            newvals = unique(newvals, noblanks=1)
            traitsStruct[range_nm] = (val_nm, newvals)
    self.traitsStruct = traitsStruct
    __traceback_info__ = ('traitsStruct', `traitsStruct`)

    pickers = self.pickers
    for i in traits:
        if REQUEST.get(i + "__picker", None):
            pickers[i] = 1
        elif pickers.has_key(i):
            del pickers[i]
    if REQUEST.get(" __requesters", None):
        pickers[" __requesters"] = 1
    self.pickers = pickers
    __traceback_info__ = ('pickers', `pickers`)

def setTrackerConfigAssignments(self, REQUEST):
    """Do setTrackerConfig operations for supporter assignments config form."""

    self.did_assignments = 1

    pickers = self.pickers
    assignments, supporters = self.assignments, self.supporters

    for s in assignments.keys():
        # Cull obsolete assignment supporters
        if s not in supporters:
            del assignments[s]

    for s in supporters:
        sa = assignments.get(s, {})
        # Cull obsolete assignment categories
        for k in sa.keys():
            if not pickers.has_key(k):
                del sa[k]
        for t in pickers.keys():
            curr = REQUEST.get(s + "__" + t, [])
            if curr and (curr[0] == " __INVERT"):
                # Inverse - represented by an empty tuple.
                curr[0] = ()
            sa[t] = curr

    self.assignments = assignments
    __traceback_info__ = ('pickers', `pickers`, 'assignments', `assignments`)
 
def addTrackerIssue(self, REQUEST=None):
    """Do permission setting and item allocations for new issue."""

    REQUEST.set('date', self.ZopeTime())

    __traceback_info__=('addTrackerIssue', `self`)
    if hasattr(self, 'Destination'):
        # We're in a factory.
        if REQUEST:
            tracker = REQUEST.PARENTS[0]
        else:
            tracker = self.Destination()
    else:
        tracker = self
    issue = tracker.TrackerIssue(REQUEST)
    issueid = tracker.add_issue(issue)
    __traceback_info__ = ('addTrackerIssue', `issueid`, `tracker.values()`)
    issue = tracker[issueid]
    requester = `REQUEST.AUTHENTICATED_USER`
    issue.requester = requester
    issue.manage_setLocalRoles(requester, ['Requester'])

    for ps in issue.propertysheets:
        ps.manage_changeProperties(REQUEST)

    # Derive issue properties and audit trail from tracker properties:
##     # Initialize the issue property sheet:
##     __traceback_info__ = (`issue.propertysheets`)
##     currentstate = PropertySheet('CurrentState')
##     for issueid, value, type in properties:
##         currentstate.manage_addProperty(issueid, value, type)
##     currentstate = FixedSchema('CurrentState', currentstate)
##      issue.propertysheets.manage_addPropertySheet('CurrentState',
##                                                   currentstate)

    traitVals = []                 # [(traitname, value), ...]
    for i in tracker.traits:
        traitVals.append(i, REQUEST.get(i + "__trait", ""))
    self.traitVals = traitVals

    issue.auditedAttrs = audit_vars('StatusVars', issue, tracker, REQUEST)

    REQUEST.set('from', requester)
    addIssueItem(issue, msgrole='request', REQUEST=REQUEST)

    issue.set_privacy(issue.private, asissue=1)

    issue.index_object()
    if REQUEST is not None:
        return REQUEST['RESPONSE'].redirect(REQUEST.URL1 + '/' + str(issueid))

def audit_vars(varrange, issue, tracker, REQUEST):
    """Package the structure of state var settings in an issue."""

    # A list of the attributes names and a dictionary going from
    # attibute names and the range of values are kept in a dictionary
    # on the issue's auditedAttrs attribute.
    if not hasattr(issue, varrange):
        raise AttributeError, ("'%s' var range property missing."
                               % varrange)
    dict = {}
    seq = []
    # Derive the data structures from the named :lines property.
    additions = []
    for var_nm in getattr(issue, varrange):
        if not var_nm:
            continue
        range_nm = var_nm + "_range"

        if hasattr(issue, range_nm):
            vals = getattr(issue, range_nm)
        elif hasattr(tracker, range_nm):
            vals = getattr(tracker, range_nm)
        else:
            raise AttributeError, ("State property '%s' not found"
                                   " (see VarInstances property '%s')"
                                   % (varrange, range_nm))
        vals = filter(None, vals)
        seq.append(var_nm)
        dict[var_nm] = vals

        # Institute the attribute and initial value on the issue:
        if hasattr(issue, var_nm):
            curval = getattr(issue, var_nm)[0]
            if type(curval) == type([]):
                curval = curval[0]
        else:
            curval = vals[0]
        setattr(issue, var_nm, curval)
        # Register the audit trail:
        additions.append((var_nm, curval))
    issue.add_state(additions, REQUEST)
    if not(hasattr(issue, 'auditedAttrs')):
        return {varrange: (seq, dict)}
    else:
        issue.auditedAttrs[varrange] = (seq, dict)
        return issue.auditedAttrs

def addIssueItem(self, msgrole='correspond', REQUEST=None):
    """Create a new item.  Raises an informative redirect for invalid input.

    The tracker rules are run here, after the item is added."""

    fromurl = (REQUEST.get('fromurl', None) or REQUEST.get('REFERER', None))

    if not (string.strip(REQUEST.get('description', ''))
            or string.strip(REQUEST.get('title', ''))):
        errmsg = ('<h3><font color="red">Issue must have a non-empty'
                  ' title or description.</font></h3>'
                  'Hit back to return to the form.')
        raise 'Redirect', errmsg

    REQUEST.set('date', self.ZopeTime())

    attachments, errmsg = snarfAttachments(REQUEST)
    if errmsg:
        errmsg = ('<h3><font color="red">'
                  + errmsg
                  + '</font></h3>'
                  + "Hit back to return to the form.")
        raise 'Redirect', errmsg

    # Stage changes (watch out for privacy conflict)

    issue = self
    user = (REQUEST and `REQUEST.AUTHENTICATED_USER`) or ''
    (itemid, item) = issue.add_item(issue.TrackerItem(REQUEST), user)

    no_post = REQUEST.get('no_post', 0)

    if REQUEST.get('private', 0):
        item.private = 1

    item.msgrole = msgrole
    if msgrole == 'request':
        setIssueStage(item, 'pending', REQUEST, no_post=no_post)
    elif msgrole == 'accept':
        setIssueStage(item, 'accepted', REQUEST, no_post=no_post)
    elif msgrole == 'resolve':
        setIssueStage(item, 'completed', REQUEST, no_post=no_post)
    elif msgrole == 'defer':
        setIssueStage(item, 'deferred', REQUEST, no_post=no_post)
    elif msgrole != 'correspond' and not no_post:
        issue.promotes(item.id, user)
        

    for ps in item.propertysheets:
        ps.manage_changeProperties(REQUEST)
## These are relevant to the issue, not the item?
##     requester = `REQUEST.AUTHENTICATED_USER`
##     item.manage_setLocalRoles(requester, ['Requester'])
                      
    origin = REQUEST.get("in_reply_to", '')
    if origin:
        try:
            origin_item = issue[origin]
            psheet = origin_item.propertysheets.Configuration
            replyprop = filter(None, psheet.getProperty('replies'))
            psheet.manage_changeProperties({'replies':
                                            replyprop + [itemid]})
        except IndexError:
            pass

    # Attachments stuff:
    for aid, origtitle, title, file in attachments:
        if not origtitle:
            title = aid                 # Avoid showing the file whole path.
        item.manage_addFile(aid, file, title)

    if int(itemid) in issue.promotes():
        was_private = item.private
        item.set_privacy(issue.private, asissue=1)
        item.private = item.private
    else:
        item.set_privacy(item.private, asissue=0)

    issue.execute_rules()


    if not no_post:
        msgid = '[t1[%s](%s)]' % (issue.id, itemid)
        send_posting(item, user, 'klm@splat.digicool.com')
##        item.messagePosts(
##                           msg_to='klm',
##                           msg_subject='%s title' % msgid,
##                           msg_from=user,
##                           msg_sender='tracker-admin@digicool.com', # XXX
##                           msg_body=REQUEST.get('description', ''),
##                           msg_response_url=item.absolute_url(),
##                           msg_ref_url=issue.absolute_url())

    if REQUEST is not None:
        return REQUEST['RESPONSE'].redirect(REQUEST.URL1)

def specialItemKnobs(self, REQUEST=None):
    """Promote or demote an item."""
    if REQUEST is None:
        return
    if REQUEST.get('new', None) == "Create a New Message":
        REQUEST.set('newitem', 1)
        return REQUEST['RESPONSE'].redirect(self.absolute_url() + "?newitem=1")
    requester = `REQUEST.AUTHENTICATED_USER`
    msgrole = REQUEST.get('msgrole', 'correspond')

    if REQUEST.get('show_knobs', None) is not None:
        return REQUEST['RESPONSE'].redirect(self.absolute_url()
                                            + "?special_knobs=1")

    if REQUEST.get('hide_knobs', None) is not None:
        return REQUEST['RESPONSE'].redirect(self.absolute_url())

    # Privacy stuff:
    privacy = REQUEST.get('privacy', 0)
    if privacy:
        if (privacy == "private") and not self.private:
            if int(self.id) in self.promotes():
                raise 'Bad Request', "Cannot conceal summary items."
            set_privacy(self, 1, asissue=0)
        elif (privacy == "public") and self.private:
            set_privacy(self, 0, asissue=0)

    if REQUEST.get('privacy_only', None) is not None:
        return REQUEST['RESPONSE'].redirect(self.absolute_url()
                                            + "?special_knobs=1")

    direction = REQUEST.get('promoting', None)
    if direction is not None:
        if direction == "Promote":
            if self.private:
                raise 'Bad Request', "Summary items cannot be private."
            self.msgrole = msgrole
            # Adjust privacy congruent with issue as a whole.
            set_privacy(self, self.aq_parent.private, asissue=1)
            self.private=0
            self.promotes(self.id, requester)
        elif direction == "Demote":
            self.msgrole = msgrole
            self.promotes(self.id, requester, demote=1)
        else:
            raise 'Bad Request', ("Unrecognized item promotion, "
                                  + `direction`)

    return REQUEST['RESPONSE'].redirect(self.absolute_url()
                                        + "?special_knobs=1")

def setIssueStage(self, stage, REQUEST, no_post=0):
    """Do issue stage-transition business, from items and issues.

    Optional no_post means don't promote the item to the summary, even for
    summary-worthy actions."""

    if stage not in self.stage_range:
        raise 'Bad Request', "Unrecognized item stage %s" % stage
    user = (REQUEST and `REQUEST.AUTHENTICATED_USER`) or ''
    tracker, issue, item = get_elements(self)
    if not issue:
        raise 'Bad Request', "No containing issue found."
    if stage == 'accepted':
        # Clear other issue owners, add this one.
        clear_local_role('IssueOwner', issue)
        issue.manage_addLocalRoles(user, ['IssueOwner'])
        if item:
            item.msgrole = 'accept'
    if stage == 'pending':
        clear_local_role('IssueOwner', issue)
        if item:
            msgrole = item.msgrole
            if item.msgrole == 'accept':
                item.msgrole = 'correspond'
    if (item is not None) and not no_post:
        if item.private:
                raise 'Bad Request', ("Action summary items cannot be private."
                                      '\n - either select "Don\'t post" or'
                                      " submit a non-action message.")
        issue.promotes(item.id, user)

    REQUEST.set('stage', stage)
    reconcileState(issue, REQUEST)

def snarfAttachments(REQUEST):
    attachments = []
    errmsg = ""
    for i in range(int(REQUEST.get('slots', 1))):
        t = REQUEST.get('file%stitle' % i, None)
        f = REQUEST.get('file%s' % i, None)
        if t or (f and f.filename):
            if t and not (f and f.filename):
                if errmsg: errmsg = errmsg + "; "
                errmsg = errmsg + ("Issue submitted with file title '%s'"
                                   " but no file." % t)
            id, title = OFS.Image.cookId(None, t, f)
            attachments.append((id, t, title, f))
    return attachments, errmsg

def reconcileState(self, REQUEST=None):
    """Set state vars, reconciling audit trail, etc."""
    if REQUEST is not None:
        # Attribute set:
        #  {'setName1': ([attr1, ...], {attr1: [val1, ...], ...}),
        #   'setName2': ...}
        for attrSet in self.auditedAttrs.values():
            for attrName in attrSet[0]:
                if (hasattr(REQUEST, attrName)
                    and (getattr(self, attrName) != REQUEST.get(attrName))):
                    setattr(self, attrName, REQUEST.get(attrName))
    return self.register_state_changes(REQUEST)

# ---------------------- Privacy. --------------------- #
# (elementEmblem has a lot - too much - of privacy stuff wrapped in...)

def elementEmblem(self, issue_id=None, item_id=None, nolink=0, REQUEST=None):
    """Return a terse link specifying a particular issue, item, or tracker.

    If there is a REQUEST, and the AUTHENTICATED_USER would not be
    authorized to view the target element, then an empty string is
    returned - actually, a string with the target in a comment.

    Unless specified, we fill in issue_id and item_id when we are within
    them, so you have to use aq_parent if you want the emblem for the parent.

    Generally, this routine is intended to make it easy to get the 'emblem'
    for the current element and for related ones - siblings, containers,
    contained, etc.  That's why it's not a method on each type of element."""

### XXX The logic needs to be reconciled to use the get_elements routine.
##     tracker, issue, item = get_elements(self)
##     if issue_id: issue = None
##     if item_id: item = None

    user = (REQUEST and `REQUEST.AUTHENTICATED_USER`) or ''
    cachekey = (issue_id, item_id, user)

    candidate = self; ids = []
    tracker = issue = item = None
    # Go up the acquisition chain, gleaning the tracker and components.
    while 1:
        if not hasattr(candidate, 'IMATRACKER'):
            break
        if (hasattr(candidate, 'aq_explicit')
            and hasattr(candidate.aq_explicit, 'IMATRACKER')):
            tracker = candidate
            break
        if (not issue_id
            and hasattr(candidate, 'aq_explicit')
            and hasattr(candidate.aq_explicit, 'IMATRACKERISSUE')):
            issue, issue_id = candidate, candidate.id
        elif (not item_id
              and hasattr(candidate, 'aq_explicit')
              and hasattr(candidate.aq_explicit, 'IMATRACKERITEM')):
            item, item_id = candidate, candidate.id
        candidate = candidate.aq_parent

    if tracker is None:
        raise AttributeError, "Failed to glean tracker"

    if (issue_id is None) and (item_id is None):
        got = tracker.title or '<Untitled Issue>'
        if nolink:
            return got
        else:
            return '<a href="%s">%s</a>' % (tracker.absolute_url(), got)

    if issue_id:
        issue = tracker[issue_id]

    if item_id is None:                 # Emblem for an issue.
        got = "(%s)" % issue_id
        if issue.private:
            got = "(" + got + ")"
            if not viewable(issue, REQUEST):
                nolink = 1
                got = '<!-- ' + got + ' -->'
        if nolink: return got
        else: return '<a href="%s">%s</a>' % (issue.absolute_url(), got)
    else:                               # Emblem for an item.
        item = issue[item_id]
        got = "[%s]" % item_id
        if item.aq_explicit.private:
            got = "[" + got + "]"
            wants = rolesForPermissionOn('View', item)
            if not viewable(item, REQUEST):
                nolink = 1
                got = '[[]] <!-- ' + got + ' -->'
        if nolink: return got
        else: return '<a href="%s">%s</a>' % (item.absolute_url(), got)
        
def tracker_role(self, role, REQUEST):
    """Verify 'staff' and 'involved' abstract roles, as well as regular roles.

    We sniff up the acquisition hierarchy in the search."""

    staff_roles = ['TrackerOwner', 'IssueOwner', 'Supporter']
    involved_roles = staff_roles + ['Requester']
    roles = []
    if type(role) == type(''):
        roles = [role]
    candidates = []
    for i in roles:
        if i == "staff": candidates = candidates + staff_roles
        elif i == "involved": candidates = candidates + involved_roles
        else: candidates.append(i)
    return REQUEST.AUTHENTICATED_USER.has_role(candidates, self)

def jiggle_supporter_roles(tracker, are, were):
    """Add supporter role for new supporters, and remove for ex-supporters."""
    added, defunct = diffs(are, were)
    for i in defunct:
        clear_local_role('Supporter', tracker, user=i)
    for i in added:
        tracker.manage_addLocalRoles(i, ['Supporter'])

def issueOwners(self):
    """Return a list of the current issue's owner(s)."""
    if hasattr(self.aq_explicit, 'IMATRACKERISSUE'):
        issue = self
    else:
        tracker, issue, item = get_elements(self)
        if issue is None:
            raise 'Bad Request', "No enclosing issue found."
    got = []
    localroles = self.__ac_local_roles__ or {}
    for who, roles in localroles.items():
        if 'IssueOwner' in roles:
            got.append(who)
    return got

def clear_local_role(role, obj, user=None):
    if user is not None:
        roleitems = ((user, obj.get_local_roles_for_userid(user)),)
    else:
        roleitems = (obj.__ac_local_roles__ or {}).items()
    for who, roles in roleitems:
        roles = filter(None, roles)
        if role in roles:
            roles.remove(role)
        if roles:
            obj.manage_setLocalRoles(who, roles)
            __traceback_info__ = ('reset', `roleitems`, `roles`, `obj`)
        else:
            # XXX This should remove entry for local user but just zeroes it?
            obj.manage_delLocalRoles([who])
            __traceback_info__ = ('delete', `roleitems`, `roles`, `obj`)

def viewable(object, REQUEST):
    "True if REQUEST user has view access to object (or if REQUEST is None)."
    userobj = (((REQUEST is not None) and REQUEST.AUTHENTICATED_USER) or None)
    if not userobj:
        return None
    roles = userobj.getRolesInContext(object)
    needs = rolesForPermissionOn('View', object)
    for i in needs:
        if i in roles:
            return 1
    return 0

def set_privacy(self, restrict, asissue=0):
    """Activate or deactivate element privacy as for an issue or item.

    Privacy at the issue level means restricted to requester and staff,
    versus open to the world.  (For staff to have a private issue, they
    must submit it themselves.)

    Privacy at the item level means restricted from requester to only
    staff, versus restricted to requester and staff.

    When promoting an item as part of the issue record, it should be set
    as for an issue, using the privacy of the issue."""

    self.private = restrict
    __traceback_info__ = (`self`, `restrict`,
                          `self.rolesOfPermission('View')`)
    minimal = ['TrackerOwner', 'Supporter', 'IssueOwner']
    intimates = minimal + ['Requester']
    maximal = intimates + ['Member', 'Anonymous']

    if not restrict:
        if asissue:
            self.manage_permission('View', roles=maximal, acquire=0)
        else:
            self.manage_permission('View', roles=intimates, acquire=0)
    else:
        if asissue:
            self.manage_permission('View', roles=intimates, acquire=0)
        else:
            if self.id in self.promotes():
                raise 'Bad Request', "Promoted items must be public."
            self.manage_permission('View', roles=minimal, acquire=0)

# -------------- Miscellaneous element data interfaces. ------------- #

def itemFilter(self, seq=None, attr=None, traits=None, negate=0,
               visonly=0, REQUEST=None):
    """Return the elements with settings for attr in traits, or not being
    traits if negate is specified.  If traits is none, just having the attr 
    qualifies (or disqualifies, if negated).

    If visonly is set, include only items accessible to the user
    (only after applying all the other criteria, including negate)."""
    # XXX This may be the basis for a more encompassing refining search.
    if seq is None:
        seq = self.values()
    
    got = []
    if type(traits) == type(""):
        traits = [traits]
    for i in seq:
        qualifies = 0
        if not attr:
            qualifies = 1
        elif hasattr(i, attr):
            if not traits or (getattr(i, attr) in traits):
                qualifies = 1
        if (qualifies and not negate) or (negate and not qualifies):
            if (not visonly or viewable(i, REQUEST)):
                got.append(i)
    return got

# -------------------- Index stuff. -------------------- #

def index_object(self):
    """Site catalog indexing

    The tracker has an explicit one for the reindexing process, so that the 
    subcomponents are traversed then."""
    return None                         # XXX !!! Temporary until cataloging

    __traceback_info__ = (`self`,)
    for i in self.values():
        for j in i.values():
            if hasattr(j, '_doindex') and j._doindex:
                j.unindex_object()
                j.index_object()
        i.unindex_object()
        i.index_object()
    self.SiteIndex.indexObject(self)

def trackerOnDeleteObject(self):
    """Object delete handler.

    Tracker has an explicit one both to do its own unidexing and also cause 
    traversal of the subcomponents."""
    for i in self.values():
        for j in i.values():
            j.unindex_object()
        i.unindex_object()

# -------------------- Rules stuff. -------------------- #

CONDITION = 'condition'
ACTIONS = 'actions'

STATE = 'state'
VOLATILE = 'volatile'
CONTAINS = 'contains'
TRANSITION = 'transition'
AND = 'and'
OR = 'or'
NOT = 'not'

SET = 'set'
INCLUDE = 'include'
REMOVE = 'remove'
NOTIFY = 'notify'
ABORT = 'abort'

ADAY = 24*60*60
ADAY = 30                         # DEBUG

def setTrackerRules(self, REQUEST=None):
    """Reset the tracker rules on the fly."""
    from urllib import quote
    __traceback_info__ = (`self`, `self.set_rules`)
    phrasing = (self._rules is not None and "(re)set") or "set"
    rules = createRules()
    self.set_rules(rules)
    if REQUEST is not None:
        destmod = quote('<font color="green">Rules %s.</font>' % phrasing)
        return REQUEST['RESPONSE'].redirect(REQUEST['URL1']
                                            + '?manage_tabs_message='
                                            + destmod)
    else:
        return rules

def createRules():
    """Create the rules for an issue tracker."""
    # XXX This is an external method for convenience.
    # XXX Eventually this will be a (perhaps more genericish) default, 
    #     and there will be a ttw interface for changing and refining them.

    # See TrackerBase.Noodge.eval_contingency(), .eval_condition(), and
    # .apply_actions() for contingency structure.
    standard = [
        ( # Forbid transition from 'pending' to non-pending states.
            (CONDITION, (AND, (STATE, 'stage', 'pending'),
                              # 'pending' was set during this transaction:
                              (VOLATILE, 'stage'),
                              # More than one stage:
                              (STATE, 'stage', None, 1),
                              # Prior stage was not 'pending':
                              (NOT, ('=', (STATE, 'stage', 'pending', 1),
                                          (STATE, 'stage', None, 1))))),
            (ACTIONS, (ABORT, ('Invalid stage transition:'
                              ' non-pending to pending')))),

        (   # Initial acknowledgement.
            # Conditions: Pending and no reception alert set.
            (CONDITION, (AND,
                          (STATE, 'stage', 'pending'),
                          (NOT, (STATE,
                                 'alert', 'reception acknowledgement', 0)))),
            (ACTIONS,
             (NOTIFY, 'Requester',
              'Your %[var "aq_parent.title"]s tracker issue',
              " '%[var title]s' (%[var id]s) has been received."
              " "
              "  Supporters have been notified, it can be viewed at"
              " %[var absolute_url]s."),
             (SET, 'alert', 'reception acknowledgement'),
             (SET, 'msgfromrole', 'requester'))),
        
        (   # Ownership pending alerts.
            # Conditions: Pending and no reception alert set.
            (CONDITION, (AND,
                          (STATE, 'stage', 'pending'),
                          ('>', (STATE,), ADAY))),
            # Due - either no alerts for notifications sent, or the last
            # one was more than a day ago.
            ( (CONDITION,
               (OR, (NOT, (STATE, 'alert', 'ownership past due notice', 0)),
                    ('>', (STATE, 'alert', 'ownership past due notice', 0),
                          ADAY))),
              (ACTIONS,
               (NOTIFY, 'Supporter',
                "Issue %[var title]s - %[var id]s needs an owner!",
                "Issue %[var issueIdentity]s is over a day old"
                " and needs an owner."),
               (INCLUDE, 'due', 'ownership'),
               (SET, 'alert', 'ownership past due notice'))),
            (   # Overdue after more than two days, and the overdue
                # notice has never been sent.
                (CONDITION,
                 (AND,
                  ('>', (STATE,), 2*ADAY),
                  (NOT, (STATE,
                         'alert', 'ownership past overdue notice', 0)))),
                (ACTIONS,
                 (NOTIFY, 'TrackerOwner',
                  "Issue %[var title]s - %[var id]s is OVERDUE for ownership!",
                  "Issue %[var issueIdentity]s is over two days old"
                  " and needs an owner.  Supporters were alerted at 1 day"
                  " mark, but no dice."),
                 (SET, 'alert', 'ownership past overdue notice'),
                 (INCLUDE, 'overdue', 'ownership')))),

        (   # Clear up ownership pending state when suitable.
            (CONDITION, (NOT, (STATE, 'stage', 'pending'))),
            ( (CONDITION, (CONTAINS, 'due', 'ownership')),
              (ACTIONS, (REMOVE, 'due', 'ownership'))),
            ( (CONDITION, (CONTAINS, 'overdue', 'ownership')),
              (ACTIONS, (REMOVE, 'overdue', 'ownership')))),

        (   # Clear up rsvp pending state when suitable.
            (CONDITION, (AND, (OR, (CONTAINS, 'due', 'rsvp'),
                                   (CONTAINS, 'overdue', 'rsvp')),
                              (NOT, (STATE, 'msgfromrole', 'Requester', 0)))),
            ( (CONDITION, (CONTAINS, 'due', 'rsvp')),
              (ACTIONS, (REMOVE, 'due', 'rsvp'))),
            ( (CONDITION, (CONTAINS, 'overdue', 'rsvp')),
              (ACTIONS, (REMOVE, 'overdue', 'rsvp')))),

        (   # Alert to languishing reply.
            (CONDITION,
             # The last correspondence is from the requester:
             (AND,
              (OR, ('<',
                    (STATE, 'msgfromrole', ['Requester', 'TrackerOwner'], 0),
                    (STATE, 'msgfromrole', 'Supporter', 0)),
               ('>', (STATE, 'msgfromrole', 'Requester', 0), 3*ADAY)))),
             # Send daily alerts to issue owner, after initial grace period:
            ( (CONDITION, (OR,
                             (NOT, (STATE, 'alert', 'rsvp due', 0)),
                             ('>', (STATE, 'alert', 'rsvp due', 0), 1*ADAY))),
              (ACTIONS,
               (SET, 'alert', 'rsvp due'),
               (NOTIFY, 'IssueOwner',
                "Issue %[var title]s %[var id]s correspondence pending",
                "Issue %[var issueIdentity]s requester is owed"
                " correspondence.")
               )),
            # This has been going on too long - alert the tracker owner once:
            ( (CONDITION, (AND,
                            ('>', (STATE, 'alert', 'rsvp due', 0), 7*ADAY),
                            (NOT, (STATE, 'alert', 'rsvp overdue', 0)))),
              (ACTIONS,
               (SET, 'alert', 'rsvp overdue'),
               (NOTIFY, 'TrackerOwner',
                "Issue %[var title]s %[var id]s correspondence WAY pending",
                "Issue %[var issueIdentity]s requester has been owed"
                " correspondence for more than 7 days."))))
        ]
    return standard

# -------------------- Sendmail tools. -------------------- #

def send_posting(item, user, to, sender='tracker-owner@digicool.com'):
    server = smtplib.SMTP(item.smtphost, item.smtpport)
    realfrom = 'tracker-admin@digicool.com'

    # * Decide where to send - to vs cc recipients:
    #   - if unowned:
    #     - if from requester, send to all relevant supporters
    #     - if from supporter, send to requester and all supporters
    #   - if owned:
    #     - if from requester, send to IssueOwners
    #     - if from IssueOwner, send to requester
    #     - if from other supporter, send to requester and IssueOwners

    #   - exclude originator

    # Note attachment locations.

    issue = item.aq_parent
    if item.msgrole == "correspond":
        msgrole = ''
    else:
        msgrole = ' ' + item.msgrole
    body = (("Subject: [tr1[%s](%s)%s] %s\n"
             % (issue.id, item.id, msgrole, item.title))
            + ("To: Tracker %s Issue (%s) Participants"
               % (issue.id, issue.title))
            + "From: Portal Member %s via <%s>\n" % (user, realfrom)
            + "\n[This is a courtesy posting in a tracker issue dialogue."
            + "To followup, visit: %s ]\n" % item.absolute_url()
            + "___________________________________________\n"
            + "\n%s\n" % item.description)
    server.sendmail('tracker-admin@digicool.com', to, body)
    server.quit()


# -------------------- Miscellaneous tools. -------------------- #

def get_elements(candidate, first_only=0):
    """Return each of the containing tracker, issue, and item that we can."""
    gots = {'tracker': None, 'issue': None, 'item': None}
    while not gots['tracker']:
        if not hasattr(candidate, 'IMATRACKER'):
            break                       # We've gone too far
        for element, flag in [('tracker', 'IMATRACKER'),
                              ('issue', 'IMATRACKERISSUE'),
                              ('item', 'IMATRACKERITEM')]:
            if (hasattr(candidate, 'aq_explicit')
                and hasattr(candidate.aq_explicit, flag)):
                if first_only:
                    return candidate
                gots[element] = candidate
        candidate = candidate.aq_parent
    return (gots['tracker'], gots['issue'], gots['item'])

def diffs(l1, l2):
    """Return two lists - elements in l1 not in l2, and vice versa.
    Preserves order."""
    l1u, l1d, l2u, l2d = [], {}, [], {}
    for i in l1: l1d[i] = 0
    for i in l2: l2d[i] = 0
    for i in l1:
        if (not l2d.has_key(i)): l1u.append(i)
    for i in l2:
        if (not l1d.has_key(i)): l2u.append(i)
    return (l1u, l2u)
def unique(lst, noblanks=0):
    """Return a copy of list with all repeats removed.  Preserves order."""
    gotd = {}
    for i in lst:
        if noblanks and i is '': continue
        if not gotd.has_key(i): gotd[i] = 0
    got = []
    for i in lst:
        if gotd.has_key(i):
            got.append(i)
            del gotd[i]
    return got


def html_quote_indented(self, text, maxchars=None, maxlines=None,
                        etc='...', charsbreak='', linesbreak=''):
    """Like html_quote, + preserve indentation and don't wrap short lines."""
    from regex import match
    from string import strip, split, expandtabs
    lines = split(html_quote(expandtabs(text, 8)), '\n')
    while lines and not strip(lines[-1]):
        del lines[-1]
    got = []
    for line in lines:
        line = line + "<br>"
        if line[0] == ' ':
            pref = match(" +", line)
            got.append((pref * "&nbsp;") + line[pref:])
        else:
            got.append(line)
    while got and (not string.strip(got[-1]) or got[-1] == '<p>'):
        del got[-1]
    origlines = len(got)
    charstruncated = linestruncated = 0
    if maxlines and len(got) > maxlines:
        got = got[:maxlines]
        linestruncated = 1
    newtext = string.join(got, '\n')
    if maxchars and (len(newtext) > maxchars):
        newtext = newtext[:maxchars]
        charstruncated = 1
    if charstruncated:
        newtext = newtext + charsbreak + etc
    elif linestruncated:
        newtext = newtext + linesbreak + etc
    return newtext

def pedigree(self):
    """Investigation utility showing the base class hierachy of a class."""
    __traceback_info__ = (`self`)
    structure = pedigree_structure(self.__class__)
    got = indented(structure, depth=0)
    return ("<!--#var standard_html_header-->"
            + ("<h3>%s Instance Inherited Classes</h3>\n" % got[0])
            + "\n<em>('...' abbreviates already elaborated classes)</em>"
            + "\n<pre>\n" + string.join(got[1:], '\n') + "\n</pre>\n"
            + "<!--#var standard_html_footer-->")
    
def indented(nesting, depth=0):
    """Indicating list nesting with indentation instead of containment.

    Ellipses appreviate already elaborated classes, and single-element
    inheritance is shown on the same line as the inheritor."""
    got = []
    for i in nesting:
        if type(i) == type(''):
            got.append((depth * '  ') + i)
##         elif len(i) == 2 and len(i[1]) == 1:
##             got.append((depth+1) * '  ' + "%s %s" % (i[0], i[1][0]))
        else:
            got = got + indented(i, depth + 1)
    return got

def pedigree_structure(klass, already_seen=None):
    """Return nested lists reflecting inheritance structure of self."""
    __traceback_info__ = (`klass`)
    bases = klass.__bases__
    seen = ""
    if already_seen is None:
        already_seen = []
    elif bases and klass in already_seen:
        seen = " ..."

    if hasattr(klass, '__module__') and klass.__module__:
        got = ["%s (%s)%s" % (klass.__name__, klass.__module__, seen)]
    else:
        got = [klass.__name__ + seen]

    if not seen:
        already_seen.append(klass)
        for i in bases:
            got.append(pedigree_structure(i, already_seen))

    return got
