##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################

"""Membership base classes"""
__version__='$Revision: 1.12 $'[11:-2]

import Globals, App.Undo, socket, regex, time, md5, sys
from string import strip, split, join, upper, lower
from Globals import HTMLFile, MessageDialog
from base64 import decodestring, encodestring
from AccessControl.User import User, UserFolder
from AccessControl.PermissionRole import _what_not_even_god_should_do
from urllib import quote, unquote







class digester:
    def __init__(self, name, pw):
        enc=md5.new()
        enc.update('%s:%s' % (name, pw))
        self.token=self.to_hex(enc.digest())

    def match(self, token):
        return token==self.token

    def to_hex(self, s, join=join):
        return join(map(lambda c: '%02x' % ord(c), s), '')



class Member(User):
    """This class is intended for use a a base class for a Z Class
       to provide enhanced User-like objects."""
    
    icon='p_/User_icon'

    def __init__(self, name, password, roles, domains):
        self.name   =name
        self._pw_   =password
        self.roles  =filter(None, roles)
        self.domains=filter(None, domains)
        
    def getUserName(self):
        return self.name

    def _getPassword(self):
        return self._pw_

    def getRoles(self):
        return self.roles
    
    def getDomains(self):
        return self.domains
    
    def authenticate(self, token, request):
        domains=self.getDomains()
        digest=digester(self.name, self._pw_)
        if domains:
            return digest.match(token) and domainSpecMatch(domains, request)
        return digest.match(token)




def authenticate(user, token, request):
    domains=user.getDomains()
    digest=digester(user.getUserName(), user._getPassword())
    if domains:
        return digest.match(token) and domainSpecMatch(domains, request)
    return digest.match(token)

_d="Microsoft Data Access Internet Publishing Provider DAV"


class MemberFolder(UserFolder):
    """This class is intended for use as a base class for a Z Class
       that implements access control and member data storage using
       http cookies."""
    
    meta_type='Member Folder'
    id       ='acl_users'
    title    ='Member Folder'
    icon     ='misc_/Membership/MemberFolder_icon'

    manage_options=(
    {'label':'Contents',   'action':'manage_main'},
    {'label':'Properties', 'action':'manage_properties'},
    {'label':'Security',   'action':'manage_access'},
    {'label':'Undo',       'action':'manage_UndoForm'},
    )

    __ac_permissions__=(
        ('Manage users', ('manage_main', 'manage_users','getUserNames',
                          'getUser','getUsers')),
        ('Change permissions', ['manage_access']),
        )

    def manage_users(self, REQUEST=None):
        """ """
        return self.manage_main(self, REQUEST)

    def err_setuser(self, request, user):
        request.other['AUTHENTICATED_USER']=user

    def p_validate(self, request, auth, roles):
        parent=self.aq_parent
        if not hasattr(parent, 'aq_parent'):
            return None
        ob=parent.aq_parent
        return p_validate(ob, request, auth, roles)


    def is_ssl(self, request):
        return (request.get('HTTPS', 'off') in ('on', 'ON')) or \
               (request.get('SERVER_PORT_SECURE', '0') == '1')

    def validate(self, request, auth='', roles=None,
                 expires='Saturday, 31-Dec-05 23:59:59 GMT'):

        if roles is _what_not_even_god_should_do:
            request.response.notFoundError()            
        req_has=request.has_key
        resp=request['RESPONSE']
        parents=request.get('PARENTS', [])
        if not parents:
            parent=self.aq_parent
        else: parent=parents[0]
        nobody=self._nobody
        super =self._super

        # Check for authorization cookie
        if req_has('__ac'):
            cookie=request['__ac']
            cookie=decodestring(unquote(cookie))
            name, token=tuple(split(cookie, ':'))
            user=self.getUser(name)
            if user is not None:
                if user.authenticate(token, request):
                    user=user.__of__(self)
                    if user.allowed(parent, roles):
                        return user
            if self._isTop() and (name==super.getUserName()) and \
               authenticate(super, token, request):
                return super.__of__(self)
        
            # Try to find a user above...
            user=self.p_validate(request, auth, roles)
            if user is not None:
                return user

            # Allow anonymous if no member can be found.
            if self._isTop() and nobody.allowed(parent, roles):
                ob=nobody.__of__(self)
                return ob

            # If it has the form variables, let it go thru
            if not req_has('__ac_password'):
                resp.expireCookie('__ac', path='/')
                self.err_setuser(request, nobody)
                raise 'Login Required', self.loginForm(self, request)

        # Check for explicit login
        if req_has('__ac_name') and req_has('__ac_password'):
            name=request['__ac_name']
            password=request['__ac_password']
            token=digester(name, password).token
            user=self.getUser(name)

            if self._isTop() and (name==super.getUserName()) and \
               authenticate(super, token, request):
                user=super
            if user is not None:
                if authenticate(user, token, request):
                    user=user.__of__(self)
                    if user.allowed(parent, roles):
                        token='%s:%s' % (name, token)
                        token=encodestring(token)
                        token=quote(token)
                        if req_has('__ac_persistent'):
                            resp.setCookie('__ac_name', name, path='/',
                                           expires=expires)
                        else:
                            resp.expireCookie('__ac_name', path='/')
                        resp.setCookie('__ac', token, path='/')
                        request['__ac']=token
                        try:
                            del request.other['__ac_name']
                            del request.other['__ac_password']
                        except:
                            pass
                        return user

            # Allow anonymous if no member can be found.
            if self._isTop() and nobody.allowed(parent, roles):
                ob=nobody.__of__(self)
                return ob
                    
            user=self.p_validate(request, auth, roles)
            if user is not None:
                return user
            self.err_setuser(request, nobody)
            raise 'Login Required', self.loginForm(self, request)

        # Check for Basic Authentication

        if auth and lower(auth[:6]) != 'basic':
            name, password = tuple(split(decodestring(split(auth)[-1]),
                                         ':',
                                         1))
            # Check for superuser
            if self._isTop() and (name == super.getUserName()) and \
               super.authenticate(password, request):
                return super

            # Try to get a user
            user = self.getUser(name)
            if user is None:
                return None

            # Try to authenticate user
            token = digester(name, password).token   #XXX hack
            if not user.authenticate(token, request):
                return None

            # We need the user to be able to acquire!
            user = user.__of__(self)

            # Try to authorize now
            if user.allowed(parent, roles):
                # Authorized, now let's set their cookie
                token = digester(name, password).token
                token = '%s:%s' % (name, token)
                token = encodestring(token)
                token = quote(token)
                resp.setCookie('__ac', token, path='/')
                request['__ac'] = token
                return user
            else:
                return None

        # No cookie or login attempt
        user=self.p_validate(request, auth, roles)
        if user is not None:
            return user
        
        # Allow anonymous if no member can be found.
        if self._isTop() and nobody.allowed(parent, roles):
            ob=nobody.__of__(self)
            return ob
        self.err_setuser(request, nobody)

        raise 'Unauthorized'

    def __bobo_traverse__(self, REQUEST, name=None):
        # Handle traversal, just to allow access to user property
        # sheets via the management interface :(
        try: return self.getUser(name).__of__(self)
        except: pass
        if hasattr(self, name):
            return getattr(self, name)
        raise 'Not Found', 'The requested resource was not found.'
    



Globals.default__class_init__(MemberFolder)
Globals.default__class_init__(Member)


def p_validate(ob, request, auth, roles):
    # This function climbs the acquisition hierarchy trying
    # to validate against any __allow_groups__ objects found.
    # This way, a UserFolder can attempt to return a valid
    # user if one exists above, but can override the default
    # authorization action (e.g. form vs. auth prompt) if no
    # user is found in higher-level objects.
    if hasattr(ob, '__allow_groups__'):
        user=ob.__allow_groups__.validate(request, auth, roles)
        if user is not None:
            return user
    if not hasattr(ob, 'aq_parent'):
        return None
    return p_validate(ob.aq_parent, request, auth, roles)


## addForm=HTMLFile('add', globals())

## def add(self, conn_id, REQUEST=None):
##     """ """
##     if hasattr(self.aq_base, 'acl_users'):
##         return MessageDialog(
##                title  ='Item Exists',
##                message='This object already contains a User Folder',
##                action ='%s/manage_main' % REQUEST['PARENT_URL'])
##     ob=SubscriptionDb(conn_id)
##     self._setObject('acl_users', ob)
##     self.__allow_groups__=self.acl_users
##     if REQUEST: return self.manage_main(self,REQUEST,update_menu=1)



###################
    
# Utility functions
###################

addr_match=regex.compile('[0-9\.\*]*').match
host_match=regex.compile('[A-Za-z0-9\.\*]*').match


def domainSpecValidate(spec):
    for ob in spec:
        sz=len(ob)
        if not ((addr_match(ob) == sz) or (host_match(ob) == sz)):
            return 0
    return 1


def domainSpecMatch(spec, request):

    if request.has_key('REMOTE_HOST'):
        host=request['REMOTE_HOST']
    else: host=''

    if request.has_key('REMOTE_ADDR'):
        addr=request['REMOTE_ADDR']
    else: addr=''

    if not host and not addr:
        return 0

    if not host:
        host=socket.gethostbyaddr(addr)[0]
    if not addr:
        addr=socket.gethostbyname(host)

    _host=split(host, '.')
    _addr=split(addr, '.')
    _hlen=len(_host)
    _alen=len(_addr)
    
    for ob in spec:
        sz=len(ob)
        _ob=split(ob, '.')
        _sz=len(_ob)

        if addr_match(ob)==sz:
            if _sz != _alen:
                continue
            fail=0
            for i in range(_sz):
                a=_addr[i]
                o=_ob[i]
                if (o != a) and (o != '*'):
                    fail=1
                    break
            if fail:
                continue
            return 1

        if host_match(ob)==sz:
            if _hlen < _sz:
                continue
            elif _hlen > _sz:
                _item=_host[-_sz:]
            else:
                _item=_host
            fail=0
            for i in range(_sz):
                h=_item[i]
                o=_ob[i]
                if (o != h) and (o != '*'):
                    fail=1
                    break
            if fail:
                continue
            return 1
    return 0



def absattr(attr):
    if callable(attr): return attr()
    return attr

def reqattr(request, attr):
    try:    return request[attr]
    except: return None


_login_html="""<html>
<head>
<title>Login</title>
</head>
<body bgcolor="#FFFFFF" link="#000099" vlink="#555555">
<p>
<form action="<!--#var SCRIPT_NAME--><!--#var PATH_INFO--><!--#if
 QUERY_STRING-->?<!--#var QUERY_STRING--><!--#/if-->" method="POST">
<table>
<tr>
  <td align="left" valign="top">
  <strong>Name</strong>
  </td>
  <td align="left" valign="top">
  <input type="TEXT" name="__ac_name" size="20">
  </td>
</tr>
<tr>
  <td align="left" valign="top">
  <strong>Password</strong>
  </td>
  <td align="left" valign="top">
  <input type="PASSWORD" name="__ac_password" size="20">
  </td>
</tr>
<tr>
  <td align="left" valign="top">
  </td>
  <td align="left" valign="top">
  <input type="SUBMIT" name="submit" value=" Ok ">
  </td>
</tr>
</table>
</form>
</body>
</html>"""


_logout_html="""<html>
<head>
<title>Logout</title>
</head>
<body bgcolor="#FFFFFF" link="#000099" vlink="#555555">
<p>
<center>
You have been logged out of the system.
</center>
</body>
</html>"""
